import logging
import copy

import ymp
from ymp.stage.stage import Stage
from ymp.stage.groupby import GroupBy
from ymp.exceptions import YmpStageError
from ymp.snakemake import ExpandLateException

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def norm_wildcards(pattern):
    for n in ("{target}", "{source}", "{:target:}"):
        pattern = pattern.replace(n, "{sample}")
    for n in ("{:targets:}", "{:sources:}"):
        pattern = pattern.replace(n, "{:samples:}")
    return pattern


def find_stage(name):
    cfg = ymp.get_config()
    registry = Stage.get_registry()

    if name.startswith("group_"):
        return GroupBy(name)
    if name.startswith("ref_"):
        refname = name[4:]
        if refname in cfg.ref:
            return cfg.ref[refname]
        else:
            raise YmpStageError(f"Unknown reference '{cfg.ref[refname]}'")
    if name in cfg.projects:
        return cfg.projects[name]
    if name in cfg.pipelines:
        return cfg.pipelines[name]
    for stage in registry.values():
        if stage.match(name):
            return stage
    raise YmpStageError(f"Unknown stage '{name}'")


class StageStack(object):
    """The "head" of a processing chain - a stack of stages
    """

    used_stacks = set()

    @classmethod
    def get(cls, path, stage=None):
        """
        Cached access to StageStack

        Args:
          path: Stage path
          stage: Stage object at head of stack
        """
        cfg = ymp.get_config()
        cache = cfg.cache.get_cache(cls.__name__, itemloadfunc=StageStack)
        res = cache[path]
        if res not in cls.used_stacks:
            cls.used_stacks.add(res)
            log.info("Stage stack %s using column %s", res, res.group)
        return res

    def __str__(self):
        return self.path

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.stage})"

    def __init__(self, path, stage=None):
        self.name = path
        self.stage_names = path.split(".")
        self.stages = [find_stage(name)
                       for name in self.stage_names]

        cfg = ymp.get_config()

        # determine project
        try:
            self.project = cfg.projects[self.stage_names[0]]
        except IndexError:
            log.error("here")
            raise YmpStageError(f"No project for stage stack {path} found")

        # determine top stage
        stage_names = copy.copy(self.stage_names)
        top_stage = stage_names.pop()
        if stage:
            if not stage.match(top_stage):
                raise YmpStageError(
                    f"Internal error: {top_stage} not matched by {stage}")
        if not stage:
            stage = find_stage(top_stage)
        self.stage = stage

        # determine grouping
        self.group = getattr(stage, "group", None)
        if stage_names and stage_names[-1].startswith("group_"):
            self.group = [stage_names.pop().split("_")[1]]

        # collect inputs
        self.prevs = self.resolve_prevs()

        if self.group is None:
            groups = list(dict.fromkeys(
                group
                for p in reversed(list(self.prevs.values()))
                for group in p.group
            ))
            self.group = self.project.minimize_variables(groups)

        log.info("Stage stack %s using column %s", self, self.group)
        prevmap = dict()
        for typ, stack in self.prevs.items():
            prevmap.setdefault(str(stack), []).append(typ)
        for stack, typ in prevmap.items():
            ftypes = ", ".join(typ).replace("/{sample}", "*")
            title = stack.split(".")[-1]
            if self.stage_names.count(title)  != 1:
                title = stack
            log.info(f".. from {title}: {ftypes}")

    def resolve_prevs(self):
        inputs = self.stage.get_inputs()
        stage = self.stage
        prevs = self._do_resolve_prevs(stage, inputs, exclude_self=True)
        if inputs:
            raise YmpStageError(self._format_missing_input_error(inputs))
        return prevs

    def _format_missing_input_error(self, inputs):
        registry = Stage.get_registry()

        # Can't find the right types, try to present useful error message:
        words = []
        for item in inputs:
            words.extend((item, "--"))
            words.extend([name for name, stage in registry.items()
                          if stage.can_provide(set(item))])
            words.extend('\n')
        text = ' '.join(words)
        return f"""
        File type(s) '{" ".join(inputs)}' required by '{self.stage}'
        not found in '{self.name}'. Stages providing missing
        file types:
        {text}
        """

    def _do_resolve_prevs(self, stage, inputs, exclude_self):
        stage_names = copy.copy(self.stage_names)
        if exclude_self:
            stage_names.pop()

        prevs = {}
        while stage_names and inputs:
            path = ".".join(stage_names)
            prev_stage = find_stage(stage_names.pop())
            prev_stack = self.get(path, prev_stage)
            provides = stage.satisfy_inputs(prev_stage, inputs)
            for typ, path in provides.items():
                if path:
                    path = ".".join(stage_names) + path
                    prev_stack = self.get(path)
                prevs[typ] = prev_stack
        return prevs

    def complete(self, incomplete):
        registry = Stage.get_registry()
        cfg = ymp.get_config()
        result = []
        groups = ("group_" + name for name in self.project.variables)
        result += (opt for opt in groups if opt.startswith(incomplete))
        refs = ("ref_" + name for name in cfg.ref)
        result += (opt for opt in refs if opt.startswith(incomplete))
        for stage in registry.values():
            for name in (stage.name, stage.altname):
                if name and name.startswith(incomplete):
                    try:
                        self.get(".".join((self.path, name)))
                        result.append(name)
                    except YmpStageError:
                        pass

        return result

    @property
    def path(self):
        """On disk location of files provided by this stack"""
        return self.stage.get_path(self)

    def all_targets(self):
        return self.stage.get_all_targets(self)

    @property
    def defined_in(self):
        return None

    def prev(self, args=None, kwargs=None):
        """
        Directory of previous stage
        """
        if not kwargs or "wc" not in kwargs:
            raise ExpandLateException()

        item = kwargs.get('item')
        _, _, suffix = kwargs['item'].partition("{:prev:}")
        suffix = norm_wildcards(suffix)

        return self.prevs[suffix]

    @property
    def targets(self):
        """
        Returns the current targets
        """
        return self.project.get_ids(self.group)

    def target(self, args, kwargs):
        """Finds the target in the prev stage matching current target"""
        prev_stage = self.prev(args, kwargs)
        prev = self.get(prev_stage.name)
        cur_target = kwargs['wc'].target
        target = self.project.get_ids(prev.group, self.group, cur_target)
        return target



import logging
import re

from ymp.exceptions import YmpRuleError
from ymp.snakemake import ColonExpander, ExpandLateException
from ymp.string import PartialFormatter
from ymp.stage.stage import Stage
from ymp.stage.stack import StageStack


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class StageExpander(ColonExpander):
    """
    - Registers rules with stages when they are created
    """
    def expand_ruleinfo(self, rule, item, expand_args, rec):
        if not Stage.active:
            return item
        stage = Stage.active

        stage._add_rule(rule)
        if not item.conda_env and stage.conda_env:
            item.conda_env = stage.conda_env

        if stage.params:
            if not item.params:
                item.params = ((), {})
            for param in stage.params:
                item.params[1][param.name] = param.param_func()

        return super().expand_ruleinfo(rule, item, expand_args, rec)

    def expand_str(self, rule, item, expand_args, rec, cb):
        if cb:
            stage = Stage.active
            Stage.active = rule.ymp_stage
        expand_args['item'] = item
        val = super().expand_str(rule, item, expand_args, rec, cb)
        if cb:
            Stage.active = stage
        return val

    def expands_field(self, field):
        return field not in 'func'

    class Formatter(ColonExpander.Formatter, PartialFormatter):
        # Careful here, TypeErrors are caught and hidden
        def get_value(self, key, args, kwargs):
            try:
                return self.get_value_(key, args, kwargs)
            except Exception as e:
                if False and not isinstance(e, ExpandLateException):
                    log.debug(f"Failed to get value: "
                              f"key={key} args={args} kwargs={kwargs}",
                              exc_info=True)
                raise

        def get_value_(self, key, args, kwargs):
            stage = Stage.active
            if "(" in key:
                args = list(args)
                if key[-1] != ")":
                    raise YmpRuleError(
                        stage,
                        f"Malformed YMP expansion string:'{key}'")
                key, _, args_str = key[:-1].partition("(")
                for arg_str in args_str.split(","):
                    try:
                        arg = int(arg_str)
                    except ValueError:
                        try:
                            arg = float(arg_str)
                        except ValueError:
                            arg = arg_str
                    if isinstance(arg, str) and arg[0] not in ('"', '"'):
                        if "wc" not in kwargs:
                            raise ExpandLateException()
                        arg = getattr(kwargs['wc'], arg)
                    args += [arg]
                
            # Check Stage variables first. We can do that always:
            if hasattr(stage, key):
                val = getattr(stage, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val

            # Check StageStack next. This requires the wildcards:
            if "wc" not in kwargs:
                raise ExpandLateException()
            wc = kwargs['wc']
            stack = StageStack.get(stage.wc2path(wc), stage)
            if hasattr(stack, key):
                val = getattr(stack, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val

            # Lastly, check the project:
            if hasattr(stack.project, key):
                val = getattr(stack.project, key)
                if hasattr(val, "__call__"):
                    val = val(args, kwargs)
                if val is not None:
                    return val
            return super().get_value(key, args, kwargs)



"""
YMP processes data in stages, each of which is contained in its own directory.

.. code-block:: snakemake

  with Stage("trim_bbmap") as S:
    S.doc("Trim reads with BBMap")
    rule bbmap_trim:
      output: "{:this:}/{sample}{:pairnames:}.fq.gz"
      input:  "{:prev:}/{sample}{:pairnames:}.fq.gz"
      ...

"""

from ymp.stage.stack import StageStack, find_stage

from ymp.stage.base import BaseStage, ConfigStage
from ymp.stage.stage import Stage
from ymp.stage.pipeline import Pipeline
from ymp.stage.reference import Reference
from ymp.stage.project import Project
from ymp.stage.expander import StageExpander
from ymp.stage.groupby import GroupBy



import copy
import logging
import re

from typing import Dict, List, Set

from snakemake.rules import Rule

from ymp.snakemake import WorkflowObject
from ymp.stage.base import BaseStage
from ymp.exceptions import YmpRuleError, YmpException


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def norm_wildcards(pattern):
    for n in ("{target}", "{source}", "{:target:}"):
        pattern = pattern.replace(n, "{sample}")
    for n in ("{:targets:}", "{:sources:}"):
        pattern = pattern.replace(n, "{:samples:}")
    return pattern


class Stage(WorkflowObject, BaseStage):
    """
    Creates a new stage

    While entered using ``with``, several stage specific variables
    are expanded within rules:

    * ``{:this:}`` -- The current stage directory
    * ``{:that:}`` -- The alternate output stage directory
    * ``{:prev:}`` -- The previous stage's directory

    """

    active = None
    """Currently active stage ("entered")"""

    def __init__(self, name: str, altname: str=None,
                 env: str=None, doc: str=None) -> None:
        """
        Args:
            name: Name of this stage
            altname: Alternate name of this stage (used for stages with
                multiple output variants, e.g. filter_x and remove_x.
            doc: See `Stage.doc`
            env: See `Stage.env`
        """
        super().__init__(name)
        # Alternate stage name
        self.altname: str = altname
        self.register()
        # Rules in this stage
        self.rules: List[Rule] = []
        # Inputs required by stage
        self._inputs: Set[str] = set()
        self._outputs: Set[str] = set()
        # Stage Parameters
        self.params: List[Param] = []
        self.requires = None
        # Regex matching self
        self._regex = None

        self.doc(doc or "")
        self.env(env)

    def env(self, name: str) -> None:
        """Add package specifications to Stage environment

        Note:
          This sets the environment for all rules within the stage,
          which leads to errors with Snakemake rule types
          not supporting conda environments

        Args:
          name: Environment name or filename

        >>> Env("blast", packages="blast =2.7*")
        >>> with Stage("test") as S:
        >>>    S.env("blast")
        >>>    rule testing:
        >>>       ...

        >>> with Stage("test", env="blast") as S:
        >>>    rule testing:
        >>>       ...

        >>> with Stage("test") as S:
        >>>    rule testing:
        >>>       conda: "blast"
        >>>       ...
        """
        self.conda_env = name

    def __enter__(self):
        if Stage.active is not None:
            raise YmpRuleError(
                self,
                f"Failed to enter stage '{self.name}', "
                f"already in stage {self.active.name}'."
            )

        Stage.active = self
        return self

    def __exit__(self, *args):
        Stage.active = None

    def __str__(self):
        if self.altname:
            return "|".join((self.name, self.altname))
        return self.name

    def __repr__(self):
        return (f"{self.__class__.__name__} {self!s} "
                f"({self.filename}:{self.lineno})")

    def _add_rule(self, rule):
        rule.ymp_stage = self
        # FIXME: disabled because it breaks pickling of Stage
        # self.rules.append(rule)

    def add_param(self, key, typ, name, value=None, default=None):
        """Add parameter to stage

        Example:
            >>> with Stage("test") as S
            >>>   S.add_param("N", "int", "nval", default=50)
            >>>   rule:
            >>>      shell: "echo {param.nval}"

            This would add a stage "test", optionally callable as "testN123",
            printing "50" or in the case of "testN123" printing "123".

        Args:
          char: The character to use in the Stage name
          typ:  The type of the parameter (int, flag)
          param: Name of parameter in params
          value: value ``{param.xyz}`` should be set to if param given
          default: default value for ``{{param.xyz}}`` if no param given
        """
        if typ == 'flag':
            self.params.append(ParamFlag(self, key, name, value, default))
        elif typ == 'int':
            self.params.append(ParamInt(self, key, name, value, default))
        elif typ == 'choice':
            self.params.append(ParamChoice(self, key, name, value, default))
        else:
            raise YmpRuleError(self, f"Unknown Stage Parameter type '{typ}'")

    def require(self, **kwargs):
        """Override inferred stage inputs

        In theory, this should not be needed. But it's simpler for now.
        """
        self.requires = kwargs

    @property
    def outputs(self):
        return self._outputs

    def get_inputs(self):
        if not self.requires:
            return copy.copy(self._inputs)
        return copy.copy(self.requires)

    def satisfy_inputs(self, other_stage, inputs) -> Dict[str, str]:
        if not self.requires: # inputs is set
            provides = other_stage.can_provide(inputs)
            # warning: `inputs -= provides.keys()` would work, but would
            # create a new object, rather than modify the set we
            # were passed.
            inputs -= set(list(provides.keys()))
            return provides

        provides = dict()
        keys = set()
        for key, input_alts in inputs.items():
            for input_alt in input_alts:
                have = other_stage.can_provide(set(
                    "/{{sample}}.{}".format(ext) for ext in input_alt
                ))
                if len(have) == len(input_alt):
                    have_new = {output: path
                                for output, path in have.items()
                                if output not in provides}
                    provides.update(have_new)
                    keys.add(key)
                    break
        for key in keys:
            del inputs[key]
        return provides

    def wc2path(self, wc):
        wildcards = self.wildcards()
        for p in self.params:
            wildcards = wildcards.replace(p.constraint, "")
        return wildcards.format(**wc)

    def match(self, name):
        if not self._regex:
            if self.altname:
                sname = "(" + "|".join((self.name, self.altname)) + ")"
            else:
                sname = self.name
            pat = sname + "".join(p.regex for p in self.params)
            self._regex = re.compile(pat)
        return self._regex.fullmatch(name) is not None

    def prev(self, args, kwargs):
        """Gathers {:prev:} calls from rules"""
        prefix, _, suffix = kwargs.get('item').partition("{:prev:}")
        if not prefix:
            self._inputs.add(norm_wildcards(suffix))
        return None

    def this(self, args=None, kwargs=None):
        """
        Directory of current stage
        """
        if not Stage.active:
            raise YmpException(
                "Use of {:this:} requires active Stage"
            )
        prefix, _, suffix = kwargs.get('item').partition("{:this:}")
        if not prefix:
            self._outputs.add(norm_wildcards(suffix))

        item = kwargs.get('item')
        if item is None:
            raise YmpException("Internal Error")
        prefix, _, pattern = item.partition("{:this:}")

        return self.wildcards(args=args, kwargs=kwargs)

    def that(self, args=None, kwargs=None):
        """
        Alternate directory of current stage

        Used for splitting stages
        """
        if not Stage.active:
            raise YmpException(
                "Use of {:that:} requires active Stage"
            )
        if not Stage.active.altname:
            raise YmpException(
                "Use of {:that:} requires with altname"
            )
        show_constraint = kwargs and kwargs.get('field') != 'input'
        return "".join(["{_YMP_DIR}", self.altname] +
                       [p.pattern(show_constraint) for p in self.params])

    def wildcards(self, args=None, kwargs=None):
        show_constraint = kwargs and kwargs.get('field') != 'input'
        return "".join(["{_YMP_DIR}", self.name] +
                       [p.pattern(show_constraint) for p in self.params])




class Param(object):
    """Stage Parameter (base class)"""
    def __init__(self, stage, key, name, value=None, default=None):
        self.stage = stage
        self.key = key
        self.name = name
        self.value = value
        self.default = default

        self.wildcard = f"_yp_{self.name}"

    @property
    def constraint(self):
        if self.regex:
            return "," + self.regex
        return ""

    def pattern(self, show_constraint=True):
        """String to add to filenames passed to Snakemake

        I.e. a pattern of the form ``{wildcard,constraint}``
        """
        if show_constraint:
            return f"{{{self.wildcard}{self.constraint}}}"
        else:
            return f"{{{self.wildcard}}}"


class ParamFlag(Param):
    """Stage Flag Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.value:
            self.value = self.key

        self.regex = f"((?:{self.key})?)"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            if getattr(wildcards, self.wildcard, None):
                return self.value
            else:
                return self.default or ""
        return name2param


class ParamInt(Param):
    """Stage Int Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default is None:
            raise YmpRuleError(
                self.stage,
                f"Stage Int Parameter must have 'default' set")

        self.regex = f"({self.key}\d+|)"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            val = getattr(wildcards, self.wildcard, None)
            if val:
                return val[len(self.key):]
            else:
                return self.default
        return name2param


class ParamChoice(Param):
    """Stage Choice Parameter"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default is not None:
            self.value += [""]
        self.regex = f"({self.key}({'|'.join(self.value)}))"

    def param_func(self):
        """Returns function that will extract parameter value from wildcards"""
        def name2param(wildcards):
            val = getattr(wildcards, self.wildcard, None)
            if val:
                return val[len(self.key):]
            else:
                return self.default
        return name2param



import logging
import os

from typing import Set, Dict, Union, List


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class BaseStage(object):
    """Base class for stage types"""
    #: The name of the stamp file that is touched to indicate
    #: completion of the stage.
    STAMP_FILENAME = "all_targets.stamp"
    def __init__(self, name: str) -> None:
        #: The name of the stage is a string uniquely identifying it
        #: among all stages.
        self.name = name

        #: The docstring describing this stage. Visible via ``ymp
        #:stage list`` and in the generated sphinx documentation.
        self.docstring: str = None

    def __str__(self) -> str:
        """Cast to string we just emit our name"""
        return self.name

    def __repr__(self):
        """Using `repr()` we emit the subclass as well as our name"""
        return f"{self.__class__.__name__}({self!s})"

    def doc(self, doc: str) -> None:
        """Add documentation to Stage

        Args:
          doc: Docstring passed to Sphinx
        """
        #: Docstring describing stage
        self.docstring = doc

    def match(self, name: str) -> bool:
        """Check if the ``name`` can refer to this stage

        As component of a `StageStack`, a stage may be identified by
        alternative names and may also be parametrized by suffix
        modifiers. Stage types supporting this behavior must override
        this function.

        """
        return name == self.name

    def get_inputs(self) -> Set[str]:
        """Returns the set of inputs required by this stage

        This function must return a copy, to ensure internal data is
        not modified.

        """
        return set()

    @property
    def outputs(self) -> Union[Set[str], Dict[str, str]]:
        """Returns the set of outputs this stage is able to generate.

        May return either a `set` or a `dict` with the dictionary
        values representing redirections in the case of virtual stages
        such as `Pipeline` or `Reference`.

        """
        return set()

    def can_provide(self, inputs: Set[str]) -> Dict[str, str]:
        """Determines which of ``inputs`` this stage can provide.

        Returns a dictionary with the keys a subset of ``inputs`` and
        the values identifying redirections. An empty string indicates
        that no redirection is to take place. Otherwise, the string is
        the suffix to be appended to the prior `StageStack`.

        """
        return {
            output: ""
            for output in inputs.intersection(self.outputs)
        }

    def get_path(self, stack: "StageStack") -> str:
        """On disk location for this stage given ``stack``.

        Called by `StageStack` to determine the real path for
        virtual stages (which must override this function).
        """
        return stack.name

    def get_all_targets(self, stack: "StageStack") -> List[str]:
        """Targets to build to complete this stage given ``stack``.

        Typically, this is the StageStack's path appended with the
        stamp name.
        """
        return [os.path.join(stack.path, self.STAMP_FILENAME)]


class ConfigStage(BaseStage):
    """Base for stages created via configuration

    These Stages derive from the ``yml.yml`` and not from a rules file.
    """
    def __init__(self, name: str, cfg: 'MultiProxy'):
        #: Semi-colon separated list of file names defining this Stage.
        self.filename = ';'.join(cfg.get_files())
        #: Line number within the first file at which this Stage is defined.
        self.lineno = next(iter(cfg.get_linenos()), None)
        super().__init__(name)
        #: The configuration object defining this Stage.
        self.cfg = cfg

    @property
    def defined_in(self):
        """List of files defining this stage
        
        Used to invalidate caches.
        """
        return self.cfg.get_files()



"""Pipelines Module

Contains classes for pre-configured pipelines comprising multiple
stages.
"""

import logging
import os

from typing import Dict, List, Set

from ymp.stage import StageStack, find_stage
from ymp.stage.base import ConfigStage
from ymp.exceptions import YmpConfigError


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Pipeline(ConfigStage):
    """
    A virtual stage aggregating a sequence of stages, i.e. a pipeline
    or sub-workflow.

    Pipelines are configured via ``ymp.yml``.

    Example:
        pipelines:
            my_pipeline:
              - stage_1
              - stage_2
              - stage_3
    """
    def __init__(self, name: str, cfg: List[str]) -> None:
        super().__init__(name, cfg)
        self.stage_names: List[str] = cfg
        self._outputs: Dict[str, str] = None

    @property
    def outputs(self) -> Dict[str, str]:
        """The outputs of a pipeline are the sum of the outputs
        of each component stage. Outputs of stages further down
        the pipeline override those generated earlier.

        TODO: Allow hiding the output of intermediary stages.
        """
        outputs = {}
        path = ""
        for stage_name in self.stage_names:
            stage = find_stage(stage_name)
            path = ".".join((path, stage_name))
            stage_outputs = stage.outputs
            if isinstance(stage_outputs, set):
                outputs.update({output: path for output in stage_outputs})
            else:
                outputs.update(stage_outputs)
        return outputs

    def can_provide(self, inputs: Set[str]) -> Dict[str, str]:
        """Determines which of ``inputs`` this stage can provide.

        The result dictionary values will point to the "real" output.
        """
        return {
            output: path
            for output, path in self.outputs.items()
            if output in inputs
        }

    @property
    def pipeline(self):
        return "." + ".".join(self.stage_names)

    def get_path(self, stack):
        prefix = stack.name.rsplit('.',1)[0]
        return prefix + self.pipeline

    def get_all_targets(self, stack):
        targets = []
        # First add the symlink for ourselves, but only if it
        # does not exist yet, due to a bug in Snakemake 5.20.1.
        if not os.path.exists(stack.name):
            targets += [stack.name]
        # Now add the target the last part of the pipeline
        # points to.
        realstack = stack.get(self.get_path(stack))
        targets.extend(realstack.stage.get_all_targets(realstack))
        return targets




import logging
import os
import re
from hashlib import sha1
from typing import Dict, Optional, Union, Set
from collections.abc import Mapping, Sequence

from ymp.snakemake import make_rule
from ymp.util import make_local_path
from ymp.stage.base import ConfigStage

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Archive(object):
    name = None
    hash = None
    tar = None
    dirname = None
    strip_components = None
    files = None

    def __init__(self, name, dirname, tar, url, strip, files):
        self.name = name
        self.dirname = dirname
        self.tar = tar
        self.url = url
        self.strip = strip
        self.files = files

        self.hash = sha1(self.tar.encode('utf-8')).hexdigest()[:8]
        self.prefix = os.path.join(self.dirname, "_unpacked_" + self.hash)

    def get_files(self):
        if isinstance(self.files, Sequence):
            return {fn: os.path.join(self.prefix, fn)
                    for fn in self.files}
        elif isinstance(self.files, Mapping):
            return {fn_ymp: os.path.join(self.prefix, fn_arch)
                    for fn_ymp, fn_arch in self.files.items()}
        else:
            raise Exception("unknown data type for reference.files")

    def make_unpack_rule(self, baserule: 'Rule'):
        docstr_tpl = """
        Unpacks {} archive:

        URL: {}

        Files:
        """

        item_tpl = """
        - {}
        """
        docstr = "\n".join([docstr_tpl.format(self.name, self.url)] +
                           [item_tpl.format(fn) for fn in self.files])
        return make_rule(
            name="unpack_{}_{}".format(self.name, self.hash),
            docstring=docstr,
            lineno=0,
            snakefile=__name__,
            parent=baserule,
            input=([], {'tar': self.tar}),
            output=([], {'files': list(self.get_files().values())}),
            params=([], {'strip': self.strip,
                         'prefix': self.prefix})
        )


class Reference(ConfigStage):
    """
    Represents (remote) reference file/database configuration
    """
    def __init__(self, name, cfg):
        super().__init__("ref_" + name, cfg)
        self.files = {}
        self.archives = []
        self.group = []
        self._outputs = None

        import ymp
        cfgmgr = ymp.get_config()

        self.dir = os.path.join(cfgmgr.dir.references, name)

        for rsc in cfg:
            if isinstance(rsc, str):
                rsc = {'url': rsc}
            self.add_files(rsc, make_local_path(cfgmgr, rsc['url']))
        self.group = self.group or ["ALL"]

    @property
    def outputs(self) -> Union[Set[str], Dict[str, str]]:
        if self._outputs is None:
            self._outputs = {
                "/" + f.replace("ALL", "{sample}"): ""
                for f in self.files
            }
        return self._outputs

    def add_files(self, rsc, local_path):
        type_name = rsc.get('type', 'fasta').lower()

        if type_name == 'fasta':
            self.files['ALL.fasta.gz'] = local_path
        elif type_name == 'fastp':
            self.files['ALL.fastp.gz'] = local_path
        elif type_name == 'gtf':
            self.files['ALL.gtf'] = local_path
        elif type_name == 'snp':
            self.files['ALL.snp'] = local_path
        elif type_name == 'tsv':
            self.files['ALL.tsv'] = local_path
        elif type_name == 'csv':
            self.files['ALL.csv'] = local_path
        elif type_name == 'dir':
            archive = Archive(name=self.name,
                              dirname=self.dir,
                              tar=local_path,
                              url=rsc['url'],
                              files=rsc['files'],
                              strip=rsc.get('strip_components', 0))
            self.files.update(archive.get_files())
            self.archives.append(archive)
        elif type_name == 'dirx':
            self.files.update({
                key: local_path + val
                for key, val in rsc.get('files', {}).items()
            })
        elif type_name == 'path':
            self.dir = rsc['url'].rstrip('/')
            self.group = rsc.get('group', [])
            for filename in os.listdir(rsc['url']):
                for regex in rsc.get('match', []):
                    match = re.fullmatch(regex, filename)
                    if not match:
                        continue
                    name = ''.join((
                        filename[:match.start('sample')],
                        'ALL',
                        filename[match.end('sample'):]
                    ))
                    self.files[name] = rsc['url'] + filename
        else:
            log.debug("unknown type {} used in reference {}"
                      "".format(type_name, self.name))


    def get_path(self, _stack):
        return self.dir

    def get_file(self, filename):
        local_path = self.files.get(filename)
        if local_path:
            return local_path
        log.error(f"{self.name}: Failed to find {filename}")
        log.warning(f"  Available: {self.files}")
        return ("YMP_FILE_NOT_FOUND__" +
                "No file {} in Reference {}"
                "".format(filename, self.name).replace(" ", "_"))

    def make_unpack_rules(self, baserule: 'Rule'):
        for archive in self.archives:
            yield archive.make_unpack_rule(baserule)

    def __str__(self):
        return os.path.join(self.dir, "ALL")
